#ifndef FEM_DATA_HPP
#define FEM_DATA_HPP

#include <fem/array.hpp>
#include <fem/str_ref.hpp>
#include <boost_adaptbx/error_utils.h>
#include <boost/shared_ptr.hpp>
#include <typeinfo>
#include <vector>

namespace fem {

  struct datum // similar to boost::any
  {
    struct placeholder
    {
      virtual ~placeholder() {}
      virtual std::type_info const& type() const = 0;
      virtual placeholder* clone() const = 0;
    };

    template<typename T>
    struct holder : placeholder
    {
      T held;

      holder(T const& value) : held(value) {}

      virtual std::type_info const&
      type() const { return typeid(T); }

      virtual placeholder*
      clone() const { return new holder(held); }

      private:
        holder& operator=(holder const&);
    };

    placeholder* content;
    mutable unsigned repeats;

    datum() : content(0), repeats(0) {}

    datum(
      char const& value)
    :
      content(new holder<char>(value)),
      repeats(1)
    {}

    datum(
      bool const& value)
    :
      content(new holder<bool>(value)),
      repeats(1)
    {}

    datum(
      int const& value)
    :
      content(new holder<int>(value)),
      repeats(1)
    {}

    datum(
      float const& value)
    :
      content(new holder<float>(value)),
      repeats(1)
    {}

    datum(
      double const& value)
    :
      content(new holder<double>(value)),
      repeats(1)
    {}

    template <typename T>
    datum(
      T const& value)
    :
      content(new holder<std::string>(std::string(value))),
      repeats(1)
    {}

    datum(
      datum const& other)
    :
       content(other.content ? other.content->clone() : 0),
       repeats(other.repeats)
    {}

    ~datum() { delete content; }

    // required for std::vector<datum> (Apple g++ 4.0.1, Visual C++ 9.0)
    // but not actually used
    datum& operator=(datum const& rhs)
    {
      throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
    }

    void
    set_repeats(
      unsigned value) const { repeats = value; }
  };

  datum const&
  operator*(
    unsigned repeats,
    datum const& value)
  {
    value.set_repeats(repeats);
    return value;
  }

  struct data_buffer
  {
    boost::shared_ptr<std::vector<datum> > objects;

    data_buffer()
    :
      objects(new std::vector<datum>)
    {}

    template <typename T>
    data_buffer&
    operator,(
      T const& val)
    {
      objects->push_back(datum(val));
      return *this;
    }
  };

  struct data_values
  {
    data_buffer values;
    size_t value_index;
    size_t repeat_index;

    data_values()
    :
      value_index(0),
      repeat_index(0)
    {}

    data_values(
      data_buffer const& values_)
    :
      values(values_),
      value_index(0),
      repeat_index(0)
    {}

    // XXX
    void
    report_types() const
    {
      size_t n = values.objects->size();
      for(size_t i=0;i!=n;i++) {
        datum const& obj = (*(values.objects))[i];
        std::cout <<
          "data_buffer[" << i << "]: " <<
            obj.content->type().name() << std::endl;
      }
    }

    datum const&
    next_datum()
    {
      datum const& result = (*(values.objects))[value_index];
      ASSERTBX(result.content != 0);
      ASSERTBX(result.repeats > 0);
      repeat_index++;
      if (repeat_index == result.repeats) {
        value_index++;
        repeat_index = 0;
      }
      return result;
    }

    data_values&
    operator,(
      bool& val)
    {
      datum const& tab_val = next_datum();
      ASSERTBX(tab_val.content->type() == typeid(bool));
      val = static_cast<datum::holder<bool>*>(tab_val.content)->held;
      return *this;
    }

    data_values&
    operator,(
      int& val)
    {
      datum const& tab_val = next_datum();
      ASSERTBX(tab_val.content->type() == typeid(int));
      val = static_cast<datum::holder<int>*>(tab_val.content)->held;
      return *this;
    }

    data_values&
    operator,(
      float& val)
    {
      datum const& tab_val = next_datum();
      ASSERTBX(tab_val.content->type() == typeid(float));
      val = static_cast<datum::holder<float>*>(tab_val.content)->held;
      return *this;
    }

    data_values&
    operator,(
      str_ref val)
    {
      datum const& tab_val = next_datum();
      ASSERTBX(tab_val.content->type() == typeid(std::string));
      std::string const& tab_str = static_cast<datum::holder<std::string>*>(
        tab_val.content)->held;
      val = tab_str.c_str();
      return *this;
    }

    template <typename T, size_t Ndims>
    data_values&
    operator,(
      array<T, Ndims>& val)
    {
      size_t n = val.size_1d();
      T* val_begin = val.begin();
      for(size_t i=0;i<n;i++) (*this), val_begin[i];
      return *this;
    }
  };

  struct values_type
  {
    template <typename T>
    data_buffer
    operator,(
      T const& val) const
    {
      return data_buffer(), val;
    }
  };

  static const values_type values = values_type();

  inline
  data_values
  data(
    data_buffer const& values)
  {
    return data_values(values);
  }

  struct data_init_flag
  {
    private:
      bool state;
      public:

    data_init_flag() : state(true) {}

    bool
    operator()()
    {
      if (!state) return false;
      state = false;
      return true;
    }
  };
}

#endif // GUARD
