#ifndef FEM_DIMENSION_HPP
#define FEM_DIMENSION_HPP

#include <fem/size_t.hpp>
#include <fem/star.hpp>
#include <limits>

namespace fem {

  template <size_t Ndims>
  struct dim_data
  {
    protected:
      ssize_t origin_[Ndims];
      size_t all_[Ndims];

      template <unsigned I>
      void
      set_origin_all_star()
      {
        origin_[I] = 1;
        all_[I] = std::numeric_limits<size_t>::max();
      }

      template <unsigned I>
      void
      set_origin_all(
        ssize_t first,
        star_type const&)
      {
        origin_[I] = first;
        all_[I] = std::numeric_limits<size_t>::max();
      }

      template <unsigned I>
      void
      set_origin_all(
        ssize_t first,
        ssize_t last)
      {
        origin_[I] = first;
        all_[I] = last - first + 1;
      }

      template <size_t SourceNdims>
      void
      copy_origin_all(
        dim_data<SourceNdims> const& source)
      {
        for(size_t i=0;i<Ndims;i++) origin_[i] = source.origin()[i];
        for(size_t i=0;i<Ndims;i++) all_[i] = source.all()[i];
      }

      public:

    ssize_t const*
    origin() const { return origin_; }

    size_t const*
    all() const { return all_; }
  };

  typedef dim_data<4> dim_buffer;

  struct dim_chain : dim_buffer
  {
    dim_chain(
      star_type const&)
    {
      this->set_origin_all_star<0>();
    }

    dim_chain(
      size_t n)
    {
      this->set_origin_all<0>(1, n);
    }

    dim_chain(
      ssize_t first,
      star_type const&)
    {
      this->set_origin_all<0>(first, star_type());
    }

    dim_chain(
      ssize_t first,
      ssize_t last)
    {
      this->set_origin_all<0>(first, last);
    }

    dim_chain&
    dim2(
      star_type const&)
    {
      this->set_origin_all_star<1>();
      return *this;
    }

    dim_chain&
    dim2(
      size_t n)
    {
      this->set_origin_all<1>(1, n);
      return *this;
    }

    dim_chain&
    dim2(
      ssize_t first,
      ssize_t last)
    {
      this->set_origin_all<1>(first, last);
      return *this;
    }

    dim_chain&
    dim3(
      star_type const&)
    {
      this->set_origin_all_star<2>();
      return *this;
    }

    dim_chain&
    dim3(
      size_t n)
    {
      this->set_origin_all<2>(1, n);
      return *this;
    }

    dim_chain&
    dim3(
      ssize_t first,
      ssize_t last)
    {
      this->set_origin_all<2>(first, last);
      return *this;
    }

    dim_chain&
    dim4(
      star_type const&)
    {
      this->set_origin_all_star<3>();
      return *this;
    }

    dim_chain&
    dim4(
      size_t n)
    {
      this->set_origin_all<3>(1, n);
      return *this;
    }

    dim_chain&
    dim4(
      ssize_t first,
      ssize_t last)
    {
      this->set_origin_all<3>(first, last);
      return *this;
    }
  };

  inline
  dim_chain
  dim1(
    star_type const&)
  {
    return dim_chain(star_type());
  }

  inline
  dim_chain
  dim1(
    size_t n)
  {
    return dim_chain(n);
  }

  inline
  dim_chain
  dim1(
    ssize_t first,
    star_type const&)
  {
    return dim_chain(first, star_type());
  }

  inline
  dim_chain
  dim1(
    ssize_t first,
    ssize_t last)
  {
    return dim_chain(first, last);
  }

  inline
  dim_buffer
  dimension(
    star_type const&)
  {
    return dim1(star_type());
  }

  inline
  dim_buffer
  dimension(
    size_t n)
  {
    return dim1(n);
  }

  inline
  dim_buffer
  dimension(
    size_t n1,
    star_type const&)
  {
    return dim1(n1).dim2(star_type());
  }

  inline
  dim_buffer
  dimension(
    size_t n1,
    size_t n2)
  {
    return dim1(n1).dim2(n2);
  }

  inline
  dim_buffer
  dimension(
    size_t n1,
    size_t n2,
    star_type const&)
  {
    return dim1(n1).dim2(n2).dim3(star_type());
  }

  inline
  dim_buffer
  dimension(
    size_t n1,
    size_t n2,
    size_t n3)
  {
    return dim1(n1).dim2(n2).dim3(n3);
  }

  inline
  dim_buffer
  dimension(
    size_t n1,
    size_t n2,
    size_t n3,
    size_t n4)
  {
    return dim1(n1).dim2(n2).dim3(n3).dim4(n4);
  }

} // namespace fem

#endif // GUARD
