#ifndef FEM_FORMAT_HPP
#define FEM_FORMAT_HPP

#include <fem/utils/char.hpp>
#include <fem/utils/string.hpp>
#include <fem/utils/token.hpp>
#include <boost/scoped_array.hpp>
#include <boost/noncopyable.hpp>
#include <vector>
#include <stdexcept>
#include <cstring>

namespace fem { namespace format {

  struct tokenizer : boost::noncopyable
  {
    protected:
      char* code;
      char* str_buf;
      unsigned stop;
      unsigned i;
      public:

    std::vector<utils::token> tokens;
    unsigned number_of_integer_tokens;

    tokenizer(
      char const* fmt,
      unsigned fmt_stop)
    {
      boost::scoped_array<char> buffer(new char[fmt_stop*2]);
      code = buffer.get();
      str_buf = code + fmt_stop;
      stop = 0;
      for(i=0;i<fmt_stop;i++) {
        char c = fmt[i];
        if (c == ' ' || c == '\t') continue;
        if (c != '\'' && c != '"') {
          code[stop++] = utils::to_lower(c);
          continue;
        }
        code[stop++] = c;
        char opening_quote = c;
        for(i++;i<fmt_stop;i++) {
          c = fmt[i];
          code[stop++] = c;
          if (c == opening_quote) {
            if (i+1 == fmt_stop) break;
            if (fmt[i+1] != c) break;
            i++;
            code[stop++] = c;
          }
        }
      }
      if (stop == 0) {
        throw std::runtime_error("Invalid FORMAT specification: empty string");
      }
      stop--;
      if (code[0] != '(') raise_invalid();
      if (code[stop] != ')') raise_invalid();
      tokens.reserve(32); // uncritical; avoids reallocation in most cases
      number_of_integer_tokens = 0;
      i = 1;
      while (i < stop) process();
      code = 0;
      str_buf = 0;
    }

    protected:
      void
      raise_invalid()
      {
        throw std::runtime_error(
          "Invalid FORMAT specification: " + std::string(code, stop+1));
      }

      bool
      starts_with(
        char const* substr,
        unsigned start)
      {
        return utils::starts_with(code, start, stop, substr);
      }

      int
      unsigned_integer_scan(
        unsigned start)
      {
        return utils::unsigned_integer_scan(code, start, stop);
      }

      void
      add_token(
        char const* type,
        unsigned start)
      {
        tokens.push_back(utils::token(type,
          std::string(&code[start], &code[i])));
      }

      void
      add_token_string()
      {
        unsigned str_size = 0;
        char opening_quote = code[i];
        for(i++;i<stop;i++) {
          char c = code[i];
          if (c == opening_quote) {
            i++;
            if (i == stop || code[i] != opening_quote) {
              tokens.push_back(utils::token("string",
                std::string(str_buf, str_size)));
              return;
            }
          }
          str_buf[str_size++] = c;
        }
        raise_invalid();
      }

      void
      process()
      {
        unsigned i_code = i;
        char c = code[i_code];
        if (c == ',') {
          i++;
          return;
        }
        if (c == 'x') {
          i++;
          add_token("format", i_code);
          return;
        }
        if (std::strchr("():/$", c) != 0) {
          i++;
          add_token("op", i_code);
          return;
        }
        if (c == '\'' || c == '"') {
          add_token_string();
          return;
        }
        if (c == '+' || c == '-') {
          i = unsigned_integer_scan(i_code+1);
          if (i < 0 || code[i] != 'p') raise_invalid();
          i++;
          add_token("format", i_code);
          return;
        }
        int j = unsigned_integer_scan(i_code);
        if (j > 0) {
          i = j;
          if (starts_with("h", j)) {
            throw std::runtime_error(
              "FATAL: Not supported: FORMAT Hollerith edit descriptor: "
                + std::string(code, stop+1));
          }
          if (starts_with("x", i) || starts_with("p", i)) {
            i++;
            add_token("format", i_code);
            return;
          }
          add_token("integer", i_code);
          number_of_integer_tokens++;
          return;
        }
        if (std::strchr("defgiz", c) != 0) {
          i++;
          j = unsigned_integer_scan(i);
          if (j > 0) {
            if (starts_with(".", j)) {
              j = unsigned_integer_scan(j+1);
              if (j < 0) raise_invalid();
            }
            i = j;
          }
          add_token("format", i_code);
          return;
        }
        if (c == 'a' || c == 'l') {
          i++;
          j = unsigned_integer_scan(i_code+1);
          if (j > 0) i = j;
          add_token("format", i_code);
          return;
        }
        if (starts_with("bn", i_code) || starts_with("bz", i_code)) {
          i += 2;
          add_token("format", i_code);
          return;
        }
        if (c == 's') {
          i++;
          if (starts_with("p", i) || starts_with("s", i)) {
            i++;
          }
          add_token("format", i_code);
          return;
        }
        if (c == 't') {
          i++;
          if (starts_with("l", i) || starts_with("r", i)) {
            i++;
          }
          j = unsigned_integer_scan(i);
          if (j < 0) raise_invalid();
          i = j;
          add_token("format", i_code);
          return;
        }
        raise_invalid();
      }
  };

  struct repeat_point
  {
    unsigned i_fmt;
    unsigned n;
    unsigned wait_for_closing_parenthesis;

    repeat_point(
      unsigned i_fmt_,
      unsigned n_)
    :
      i_fmt(i_fmt_),
      n(n_),
      wait_for_closing_parenthesis(false)
    {}
  };

}} // namespace fem::format

#endif // GUARD
