#ifndef FEM_IO_HPP
#define FEM_IO_HPP

#include <fem/inquire_chain_decl.hpp>
#include <fem/open_chain_decl.hpp>
#include <fem/utils/simple_streams.hpp>
#include <fem/utils/string.hpp>
#include <boost/noncopyable.hpp>
#include <boost/format.hpp>
#include <boost/shared_ptr.hpp>
#include <map>
#include <cstdio>

namespace fem {

  struct std_file
  {
    std::FILE* ptr;

    std_file(std::FILE* ptr_=0) : ptr(ptr_) {}
  };

  struct io : boost::noncopyable
  {
    std::map<int, std_file> unit_streams;
    std::map<int, std::string> unit_file_names;

    io()
    {
      unit_streams[0] = std_file(stderr);
      unit_streams[5] = std_file(stdin);
      unit_streams[6] = std_file(stdout);
    }

    static
    bool
    is_std_unit(
      int unit)
    {
      return (
           unit == 0
        || unit == 5
        || unit == 6);
    }

    ~io()
    {
      typedef std::map<int, std_file>::iterator it;
      it e = unit_streams.end();
      for(it i=unit_streams.begin();i!=e;i++) {
        if (is_std_unit(i->first)) continue;
        std::FILE* f = i->second.ptr;
        if (f != 0) {
          std::fclose(f);
        }
      }
    }

    inquire_chain
    inquire_unit(
      int unit)
    {
      return inquire_chain(this, unit);
    }

    inquire_chain
    inquire_file(
      std::string const& file,
      bool blank_padding_removed_already=false)
    {
      if (blank_padding_removed_already) {
        return inquire_chain(this, file);
      }
      return inquire_chain(
        this, utils::strip_leading_and_trailing_blank_padding(file));
    }

    open_chain
    open(
      int unit,
      std::string const& file,
      bool blank_padding_removed_already=false)
    {
      std_file& sf = unit_streams[unit];
      if (sf.ptr != 0) {
        std::fclose(sf.ptr);
        sf.ptr = 0;
      }
      if (blank_padding_removed_already) {
        unit_file_names[unit] = file;
      }
      else {
        unit_file_names[unit] =
          utils::strip_leading_and_trailing_blank_padding(file);
      }
      return open_chain(this, unit);
    }

    void
    close(
      int unit)
    {
      open(unit, std::string());
    }

    std::string
    file_name(
      int unit)
    {
      std::string fn = unit_file_names[unit];
      if (fn.size() == 0 && !is_std_unit(unit)) {
        fn = (boost::format("fem_io_unit_%03d") % unit).str();
        unit_file_names[unit] = fn;
      }
      return fn;
    }

    boost::shared_ptr<utils::simple_ostream>
    simple_ostream(
      int unit)
    {
      std_file& sf = unit_streams[unit];
      if (sf.ptr == 0) {
        std::string fn = file_name(unit);
        sf.ptr = std::fopen(fn.c_str(), "wb");
        if (sf.ptr == 0) {
          throw std::runtime_error("Cannot open file for writing: " + fn);
        }
      }
      return boost::shared_ptr<utils::simple_ostream>(new
        utils::simple_ostream_to_c_file(sf.ptr));
    }

    boost::shared_ptr<utils::simple_istream>
    simple_istream(
      int unit)
    {
      std_file& sf = unit_streams[unit];
      if (sf.ptr == 0) {
        std::string fn = file_name(unit);
        sf.ptr = std::fopen(fn.c_str(), "rb");
        if (sf.ptr == 0) {
          throw std::runtime_error("Cannot open file for reading: " + fn);
        }
      }
      return boost::shared_ptr<utils::simple_istream>(new
        utils::simple_istream_from_c_file(sf.ptr));
    }
  };

} // namespace fem

#endif // GUARD
