#ifndef FEM_READ_HPP
#define FEM_READ_HPP

#include <fem/common.hpp>
#include <fem/format.hpp>
#include <fem/star.hpp>
#include <fem/str_ref.hpp>
#include <fem/utils/string_to_double_fmt.hpp>
#include <boost_adaptbx/error_utils.h>

namespace fem {

  struct read_end : std::runtime_error
  {
    explicit
    read_end(std::string const& msg) : std::runtime_error(msg) {}
  };

  struct read_err : std::runtime_error
  {
    explicit
    read_err(std::string const& msg) : std::runtime_error(msg) {}
  };

  class read_loop // TODO copy-constructor potential performance problem
  {
    private:
      boost::shared_ptr<utils::simple_istream> inp;
      bool have_fmt;
      std::vector<utils::token> fmt_tokens;
      unsigned i_fmt;
      unsigned i_fmt_wrap;
      std::vector<format::repeat_point> repeat_points;
      bool blanks_zero;
      int exp_scale;

    public:

      read_loop(
        common& cmn,
        int const& unit)
      :
        inp(cmn.io.simple_istream(unit)),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0),
        blanks_zero(false),
        exp_scale(0)
      {}

      read_loop(
        common& cmn,
        int const& unit,
        star_type const&)
      :
        inp(cmn.io.simple_istream(unit)),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0),
        blanks_zero(false),
        exp_scale(0)
      {}

      read_loop(
        common& cmn,
        int const& unit,
        char const* fmt)
      :
        inp(cmn.io.simple_istream(unit)),
        have_fmt(true),
        i_fmt(0),
        i_fmt_wrap(0),
        blanks_zero(false),
        exp_scale(0)
      {
        init_fmt(fmt);
      }

      read_loop(
        str_cref const& internal_file,
        star_type const&)
      :
        inp(boost::shared_ptr<utils::simple_istream>(new
          utils::simple_istream_from_char_ptr_and_size(
            internal_file.elems(), internal_file.len()))),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0),
        blanks_zero(false),
        exp_scale(0)
      {}

      read_loop(
        str_cref const& internal_file,
        char const* fmt)
      :
        inp(boost::shared_ptr<utils::simple_istream>(new
          utils::simple_istream_from_char_ptr_and_size(
            internal_file.elems(), internal_file.len()))),
        have_fmt(true),
        i_fmt(0),
        i_fmt_wrap(0),
        blanks_zero(false),
        exp_scale(0)
      {
        init_fmt(fmt);
      }

      protected:
        void
        init_fmt(
          char const* fmt)
        {
          format::tokenizer tz(fmt, std::strlen(fmt));
          fmt_tokens.swap(tz.tokens);
          repeat_points.reserve(tz.number_of_integer_tokens);
        }
        public:

      std::string const&
      next_edit_descriptor()
      {
        while (true) {
          if (repeat_points.size() != 0) {
            format::repeat_point& rp = repeat_points.back();
            if (!rp.wait_for_closing_parenthesis) {
              if (rp.n == 0) {
                repeat_points.pop_back();
                continue;
              }
              i_fmt = rp.i_fmt;
              rp.n--;
            }
          }
          if (i_fmt == fmt_tokens.size()) {
            ASSERTBX(fmt_tokens.size() != 0); // better msg TODO
            i_fmt = i_fmt_wrap;
          }
          utils::token const& t = fmt_tokens[i_fmt++];
          std::string const& tv = t.value;
          if (t.type == "string") {
            throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
          }
          else if (t.type == "integer") {
            repeat_points.push_back(format::repeat_point(
              i_fmt,
              utils::unsigned_integer_value(tv.data(), tv.size())));
          }
          else if (t.type == "op") {
            if (tv[0] == '(') {
              if (repeat_points.size() != 0) {
                format::repeat_point& rp = repeat_points.back();
                ASSERTBX(rp.i_fmt+1 == i_fmt); // better msg TODO
                ASSERTBX(!rp.wait_for_closing_parenthesis); // better msg TODO
                rp.wait_for_closing_parenthesis = true;
                if (repeat_points.size() == 1) {
                  i_fmt_wrap = rp.i_fmt;
                }
              }
            }
            else if (tv[0] == ')') {
              if (repeat_points.size() != 0) {
                format::repeat_point& rp = repeat_points.back();
                ASSERTBX(rp.wait_for_closing_parenthesis); // better msg TODO
                rp.wait_for_closing_parenthesis = false;
              }
            }
            else if (tv[0] == ':') {
              // ignored
            }
            else if (tv[0] == '/') {
              throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
            }
            else if (tv[0] == '$') {
              throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
            }
            else {
              throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
            }
          }
          else if (t.type == "format") {
            if (utils::ends_with_char(tv, 'x')) {
              if (tv.size() == 1) {
                process_fmt_x(1);
              }
              else {
                process_fmt_x(utils::signed_integer_value(
                  tv.data(), 0, tv.size()-1));
              }
            }
            else if (std::strchr("adefgilz", tv[0]) != 0) {
              return tv;
            }
            else if (utils::ends_with_char(tv, 'p')) {
              if (tv.size() == 1) {
                exp_scale = 1;
              }
              else {
                exp_scale = utils::signed_integer_value(
                  tv.data(), 0, tv.size()-1);
              }
            }
            else if (tv[0] == 't') {
              throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
            }
            else if (tv[0] == 's') {
              throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
            }
            else if (tv[0] == 'b') {
              blanks_zero = (tv[1] == 'z');
            }
            else {
              throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
            }
          }
          else {
            throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
          }
        }
      }

      int
      inp_get()
      {
        int result = inp->get();
        if (utils::is_stream_err(result)) {
          throw read_err("Error during read");
        }
        return result;
      }

      void
      process_fmt_x(
        unsigned n)
      {
        for(unsigned i=0;i<n;i++) {
          int c = inp_get();
          if (c == utils::stream_end) {
            return;
          }
          if (utils::is_end_of_line(c)) {
            inp->backup();
            return;
          }
        }
      }

      read_loop&
      operator,(
        char& val)
      {
        throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
        return *this;
      }

      read_loop&
      operator,(
        bool& val)
      {
        throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
        return *this;
      }

      read_loop&
      operator,(
        int& val)
      {
        if (!have_fmt) {
          val = static_cast<int>(read_star_long());
        }
        else {
          std::string const& ed = next_edit_descriptor();
          int n = ed.size();
          if (ed[0] == 'i' && n > 1) {
            n = utils::unsigned_integer_value(ed.data(), 1, n);
            val = static_cast<int>(read_fmt_long(n));
          }
          else {
            val = static_cast<int>(read_star_long());
          }
        }
        return *this;
      }

      read_loop&
      operator,(
        float& val)
      {
        val = static_cast<float>(
          (have_fmt ? read_fmt_double() : read_star_double()));
        return *this;
      }

      read_loop&
      operator,(
        double& val)
      {
        val = (have_fmt ? read_fmt_double() : read_star_double());
        return *this;
      }

      read_loop&
      operator,(
        str_ref const& val)
      {
        int vl = val.len();
        int n = vl;
        if (have_fmt) {
          std::string const& ed = next_edit_descriptor();
          if (ed[0] == 'a' && ed.size() > 1) {
            n = utils::unsigned_integer_value(ed.data(), 1, ed.size());
            if (n > vl) n = vl;
          }
        }
        int i=0;
        for(;i<n;i++) {
          int c = inp_get();
          if (utils::is_stream_end(c)) {
            if (i == 0) {
              throw read_end("End of input while reading string");
            }
            break;
          }
          if (utils::is_end_of_line(c)) {
            inp->backup();
            break;
          }
          val[i] = c;
        }
        for(;i<vl;i++) val[i] = ' ';
        return *this;
      }

      template <typename T, size_t Ndims>
      read_loop&
      operator,(
        ref<T, Ndims> const& val)
      {
        throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
        return *this;
      }

      ~read_loop()
      {
        while (true) {
          int c = inp->get();
          if (   utils::is_stream_end_or_err(c)
              || utils::is_end_of_line(c)) {
            break;
          }
        }
      }

      long
      read_fmt_long(
        unsigned n)
      {
        bool had_non_blank = false;
        bool negative = false;
        long result = 0;
        for(unsigned i=0;i<n;i++) {
          int c = inp->get();
          if (utils::is_stream_end_or_err(c)) {
            if (i == 0) {
              throw read_end(
                "End-of-file while reading integer value");
            }
            break;
          }
          if (c == ',' || utils::is_end_of_line(c)) {
            break;
          }
          if (c != ' ') {
            if (!had_non_blank) {
              had_non_blank = true;
              if (c == '-') {
                negative = true;
                continue;
              }
              if (c == '+') {
                continue;
              }
            }
            if (!utils::is_digit(c)) {
              throw read_err(
                "Invalid character while reading integer value.");
            }
            result *= 10;
            result += utils::digit_as_int(c);
          }
        }
        if (negative) result *= -1;
        return result;
      }

      long
      read_star_long()
      {
        while (true) { // loop scanning for first non-whitespace
          int c = inp->get();
          if (utils::is_stream_end_or_err(c)) {
            throw read_end(
              "End of input while reading integer value");
          }
          if (!utils::is_whitespace(c)) {
            bool negative = (c == '-');
            if (negative || c == '+') {
              c = inp->get();
              if (utils::is_stream_end_or_err(c)) {
                read_end(
                  "End of input while reading integer value");
              }
            }
            long result = 0;
            while (true) { // loop collecting digits
              if (!utils::is_digit(c)) {
                read_err(
                  "Invalid character while reading integer value.");
              }
              result *= 10;
              result += utils::digit_as_int(c);
              c = inp->get();
              if (   utils::is_stream_end_or_err(c)
                  || utils::is_whitespace(c)
                  || c == ',') {
                if (negative) result *= -1;
                if (utils::is_end_of_line(c)) inp->backup();
                return result;
              }
            }
          }
        }
      }

      double
      read_fmt_double()
      {
        std::string const& ed = next_edit_descriptor();
        int n = ed.size();
        if (n < 2 || std::strchr("defg", ed[0]) == 0) {
          return read_star_double();
        }
        int iw = utils::unsigned_integer_scan(ed.data(), 1, ed.size());
        int w = utils::unsigned_integer_value(ed.data(), 1, iw);
        int d = 0;
        if (iw+1 != ed.size()) {
          d = utils::unsigned_integer_value(ed.data(), iw+1, ed.size());
        }
        utils::string_to_double_fmt conv(*inp, w, d, blanks_zero, exp_scale);
        if (conv.error_message != 0) {
          throw read_end(conv.error_message);
        }
        return conv.result;
      }

      double
      read_star_double()
      {
        utils::string_to_double conv(*inp);
        if (conv.error_message != 0) {
          throw read_end(conv.error_message);
        }
        int c = inp->get();
        if (   utils::is_stream_end_or_err(c)
            || utils::is_whitespace(c)
            || c == ',') {
          if (utils::is_end_of_line(c)) inp->backup();
          return conv.result;
        }
        throw read_err(
          "Invalid character while reading floating-point value.");
      }
  };

  struct common_read
  {
    common& cmn;

    common_read(
      common& cmn_)
    :
      cmn(cmn_)
    {}

    read_loop
    operator()(
      int unit)
    {
      return read_loop(cmn, unit);
    }

    read_loop
    operator()(
      int unit,
      star_type const&)
    {
      return read_loop(cmn, unit, star);
    }

    read_loop
    operator()(
      int const& unit,
      char const* fmt)
    {
      return read_loop(cmn, unit, fmt);
    }

    read_loop
    operator()(
      str_cref const& internal_file,
      star_type const&)
    {
      return read_loop(internal_file, star);
    }

    read_loop
    operator()(
      str_cref const& internal_file,
      char const* fmt)
    {
      return read_loop(internal_file, fmt);
    }
  };

} // namespace fem

#endif // GUARD
