#ifndef FEM_REF_HPP
#define FEM_REF_HPP

#include <fem/dimension.hpp>
#include <algorithm>

namespace fem {

  template <size_t Ndims>
  struct ref_dims_base : dim_data<Ndims>
  {
    size_t
    size_1d() const
    {
      size_t result = 1;
      for(size_t i=0;i<Ndims;i++) result *= this->all_[i];
      return result;
    }
  };

  template <size_t Ndims>
  struct ref_dims;

  template <>
  struct ref_dims<1> : ref_dims_base<1>
  {
    size_t
    index_1d(
      ssize_t i1) const
    {
      return i1 - this->origin_[0];
    }
  };

  template <>
  struct ref_dims<2> : ref_dims_base<2>
  {
    size_t
    index_1d(
      ssize_t i1,
      ssize_t i2) const
    {
      return
          (i2 - this->origin_[1]) * this->all_[0]
        + (i1 - this->origin_[0]);
    }

  };

  template <>
  struct ref_dims<3> : ref_dims_base<3>
  {
    size_t
    index_1d(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3) const
    {
      return
          ((i3 - this->origin_[2])  * this->all_[1]
        +  (i2 - this->origin_[1])) * this->all_[0]
        +  (i1 - this->origin_[0]);
    }
  };

  template <>
  struct ref_dims<4> : ref_dims_base<4>
  {
    size_t
    index_1d(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3,
      ssize_t i4) const
    {
      return
          (((i4 - this->origin_[3])  * this->all_[2]
        +   (i3 - this->origin_[2])) * this->all_[1]
        +   (i2 - this->origin_[1])) * this->all_[0]
        +   (i1 - this->origin_[0]);
    }
  };

  template <typename T, size_t Ndims=1>
  struct cref : ref_dims<Ndims>
  {
    protected:
      T const* elems_;

      cref() {}

      public:

    cref(
      T const& val)
    {
      elems_ = &val;
    }

    template <size_t OtherNdims>
    cref(
      cref<T, OtherNdims> const& other)
    {
      elems_ = other.begin();
    }

    T const*
    begin() const { return elems_; }

    void
    operator()(
      dim_buffer const& dims)
    {
      this->copy_origin_all(dims);
    }

    T const&
    operator()(
      ssize_t i1) const
    {
      return elems_[this->index_1d(i1)];
    }

    T const&
    operator()(
      ssize_t i1,
      ssize_t i2) const
    {
      return elems_[this->index_1d(i1, i2)];
    }

    T const&
    operator()(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3) const
    {
      return elems_[this->index_1d(i1, i2, i3)];
    }

    T const&
    operator()(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3,
      ssize_t i4) const
    {
      return elems_[this->index_1d(i1, i2, i3, i4)];
    }
  };

  template <typename T, size_t Ndims=1>
  struct ref : cref<T, Ndims>
  {
    protected:
      ref() {}

      public:

    ref(
      T& val)
    {
      this->elems_ = &val;
    }

    template <size_t OtherNdims>
    ref(
      ref<T, OtherNdims> const& other)
    {
      this->elems_ = other.begin();
    }

    T*
    begin() const { return const_cast<T*>(this->elems_); }

    void
    operator()(
      dim_buffer const& dims)
    {
      cref<T, Ndims>::operator()(dims);
    }

    T&
    operator()(
      ssize_t i1) const
    {
      return this->begin()[this->index_1d(i1)];
    }

    T&
    operator()(
      ssize_t i1,
      ssize_t i2) const
    {
      return this->begin()[this->index_1d(i1, i2)];
    }

    T&
    operator()(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3) const
    {
      return this->begin()[this->index_1d(i1, i2, i3)];
    }

    T&
    operator()(
      ssize_t i1,
      ssize_t i2,
      ssize_t i3,
      ssize_t i4) const
    {
      return this->begin()[this->index_1d(i1, i2, i3, i4)];
    }

    operator
    T&() const { return *(this->begin()); }

    void
    fill(
      T const& v)
    {
      std::fill_n(this->begin(), this->size_1d(), v);
    }
  };

} // namespace fem

#endif // GUARD
