#ifndef FEM_WRITE_HPP
#define FEM_WRITE_HPP

#include <fem/common.hpp>
#include <fem/format.hpp>
#include <fem/ref.hpp>
#include <fem/star.hpp>
#include <fem/str_ref.hpp>
#include <boost_adaptbx/error_utils.h>

namespace fem {

  struct write_err : std::runtime_error
  {
    explicit
    write_err(std::string const& msg) : std::runtime_error(msg) {}
  };

  class write_loop // TODO copy-constructor potential performance problem
  {
    private:
      boost::shared_ptr<utils::simple_ostream> out;
      int internal_file_len;
      unsigned pos;
      bool prev_was_string;
      bool have_fmt;
      std::vector<utils::token> fmt_tokens;
      unsigned i_fmt;
      unsigned i_fmt_wrap;
      std::vector<format::repeat_point> repeat_points;

    public:

      write_loop(
        common& cmn,
        int const& unit)
      :
        out(cmn.io.simple_ostream(unit)),
        internal_file_len(-1),
        pos(0),
        prev_was_string(false),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0)
      {}

      write_loop(
        common& cmn,
        int const& unit,
        star_type const&)
      :
        out(cmn.io.simple_ostream(unit)),
        internal_file_len(-1),
        pos(0),
        prev_was_string(false),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0)
      {}

      write_loop(
        common& cmn,
        int const& unit,
        char const* fmt)
      :
        out(cmn.io.simple_ostream(unit)),
        internal_file_len(-1),
        pos(0),
        prev_was_string(false),
        have_fmt(true),
        i_fmt(0),
        i_fmt_wrap(0)
      {
        format::tokenizer tz(fmt, std::strlen(fmt));
        fmt_tokens.swap(tz.tokens);
        repeat_points.reserve(tz.number_of_integer_tokens);
      }

      write_loop(
        str_ref const& internal_file,
        star_type const&)
      :
        out(boost::shared_ptr<utils::simple_ostream>(new
          utils::simple_ostream_to_char_ptr_and_size(
            internal_file.elems(), internal_file.len()))),
        internal_file_len(internal_file.len()),
        pos(0),
        prev_was_string(false),
        have_fmt(false),
        i_fmt(0),
        i_fmt_wrap(0)
      {}

      write_loop(
        str_ref const& internal_file,
        char const* fmt)
      :
        out(boost::shared_ptr<utils::simple_ostream>(new
          utils::simple_ostream_to_char_ptr_and_size(
            internal_file.elems(), internal_file.len()))),
        internal_file_len(internal_file.len()),
        pos(0),
        prev_was_string(false),
        have_fmt(true),
        i_fmt(0),
        i_fmt_wrap(0)
      {
        format::tokenizer tz(fmt, std::strlen(fmt));
        fmt_tokens.swap(tz.tokens);
        repeat_points.reserve(tz.number_of_integer_tokens);
      }

      std::string const&
      next_edit_descriptor(
        bool final=false)
      {
        while (true) {
          if (repeat_points.size() != 0) {
            format::repeat_point& rp = repeat_points.back();
            if (!rp.wait_for_closing_parenthesis) {
              if (rp.n == 0) {
                repeat_points.pop_back();
                continue;
              }
              i_fmt = rp.i_fmt;
              rp.n--;
            }
          }
          if (i_fmt == fmt_tokens.size()) {
            if (final) {
              static const std::string empty("");
              return empty;
            }
            ASSERTBX(fmt_tokens.size() != 0); // better msg TODO
            to_stream_fmt("\n", 1);
            i_fmt = i_fmt_wrap;
          }
          utils::token const& t = fmt_tokens[i_fmt++];
          std::string const& tv = t.value;
          if (t.type == "string") {
            to_stream_fmt(tv.data(), tv.size());
          }
          else if (t.type == "integer") {
            repeat_points.push_back(format::repeat_point(
              i_fmt,
              utils::unsigned_integer_value(tv.data(), tv.size())));
          }
          else if (t.type == "op") {
            if (tv[0] == '(') {
              if (repeat_points.size() != 0) {
                format::repeat_point& rp = repeat_points.back();
                ASSERTBX(rp.i_fmt+1 == i_fmt); // better msg TODO
                ASSERTBX(!rp.wait_for_closing_parenthesis); // better msg TODO
                rp.wait_for_closing_parenthesis = true;
                if (repeat_points.size() == 1) {
                  i_fmt_wrap = rp.i_fmt;
                }
              }
            }
            else if (tv[0] == ')') {
              if (repeat_points.size() != 0) {
                format::repeat_point& rp = repeat_points.back();
                ASSERTBX(rp.wait_for_closing_parenthesis); // better msg TODO
                rp.wait_for_closing_parenthesis = false;
              }
            }
            else if (tv[0] == ':') {
              // ignored
            }
            else if (tv[0] == '/') {
              to_stream_fmt("\n", 1);
            }
            else if (tv[0] == '$') {
              // TODO
            }
            else {
              throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
            }
          }
          else if (t.type == "format") {
            if (utils::ends_with_char(tv, 'x')) {
              unsigned n = tv.size();
              if (n != 1) n = utils::unsigned_integer_value(
                tv.data(), n-1);
              for(unsigned i=0;i<n;i++) to_stream(" ", 1);
            }
            else if (std::strchr("adefgilz", tv[0]) != 0) {
              return tv;
            }
            else if (utils::ends_with_char(tv, 'p')) {
              // TODO
            }
            else if (tv[0] == 't') {
              // TODO
            }
            else if (tv[0] == 's') {
              // TODO
            }
            else if (tv[0] == 'b') {
              // TODO
            }
            else {
              throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
            }
          }
          else {
            throw BOOST_ADAPTBX_UNREACHABLE_ERROR();
          }
        }
      }

      write_loop&
      operator,(
        char const& val)
      {
        to_stream(&val, 1, /*space*/ !prev_was_string);
        prev_was_string = true;
        return *this;
      }

      write_loop&
      operator,(
        char const* val)
      {
        if (!have_fmt) {
          to_stream_star(val, std::strlen(val), /*space*/ !prev_was_string);
          prev_was_string = true;
        }
        else {
          std::string const& ed = next_edit_descriptor();
          if (ed[0] == 'a') {
            int n = ed.size();
            ASSERTBX(n+2 < 64);
            char fmt[64];
            fmt[0] = '%';
            std::strncpy(fmt+1, ed.data()+1, n-1);
            fmt[n] = 's';
            fmt[n+1] = '\0';
            char buf[64];
            n = std::sprintf(buf, fmt, val);
            to_stream_fmt(buf, n);
          }
          else {
            to_stream_fmt(val, std::strlen(val));
          }
        }
        return *this;
      }

      write_loop&
      operator,(
        bool const& val)
      {
        to_stream((val ? "T" : "F"), 1);
        prev_was_string = false;
        return *this;
      }

      write_loop&
      operator,(
        int const& val)
      {
        if (!have_fmt) {
          char buf[64];
          int n = std::sprintf(buf, "%11d", val);
          to_stream_star(buf, n);
          prev_was_string = false;
        }
        else {
          std::string const& ed = next_edit_descriptor();
          if (ed[0] == 'i') {
            int n = ed.size();
            ASSERTBX(n+2 < 64);
            char fmt[64];
            fmt[0] = '%';
            std::strncpy(fmt+1, ed.data()+1, n-1);
            fmt[n] = 'd';
            fmt[n+1] = '\0';
            char buf[64];
            n = std::sprintf(buf, fmt, val);
            to_stream_fmt(buf, n);
          }
          else {
            char buf[64];
            int n = std::sprintf(buf, "%d", val);
            to_stream_fmt(buf, n);
          }
        }
        return *this;
      }

      write_loop&
      operator,(
        float const& val)
      {
        if (!have_fmt) {
          char buf[64];
#if defined(_MSC_VER)
          int n = std::sprintf(buf, "%15.7E", val);
          if (n == 15 && buf[12] == '0') {
            buf[12] = buf[13];
            buf[13] = buf[14];
            buf[14] = '\0';
            n = 14;
          }
#else
          int n = std::sprintf(buf, "%14.7E", val);
#endif
          to_stream(buf, n);
          prev_was_string = false;
        }
        else {
          std::string const& ed = next_edit_descriptor();
          if (ed[0] == 'f') {
            int n = ed.size();
            ASSERTBX(n+2 < 64);
            char fmt[64];
            fmt[0] = '%';
            std::strncpy(fmt+1, ed.data()+1, n-1);
            fmt[n] = 'f';
            fmt[n+1] = '\0';
            char buf[64];
            n = std::sprintf(buf, fmt, val);
            to_stream_fmt(buf, n);
          }
          else {
            char buf[64];
            int n = std::sprintf(buf, "%.6g", val);
            to_stream_fmt(buf, n);
          }
        }
        return *this;
      }

      write_loop&
      operator,(
        double const& val)
      {
        char buf[64];
        int n = std::sprintf(buf, "%23.15E", val);
        int offs = 0;
        if (n == 23) {
          char c = buf[20];
          if ((c == '+' || c == '-') && buf[0] == ' ') {
            offs = 1;
            buf[24] = '\0';
            buf[23] = buf[22];
            buf[22] = buf[21];
            buf[21] = '0';
          }
        }
        to_stream(buf+offs, n);
        prev_was_string = false;
        return *this;
      }

      write_loop&
      operator,(
        str_cref const& val)
      {
        if (!have_fmt) {
          to_stream(val.elems(), val.len(), /*space*/ !prev_was_string);
          prev_was_string = true;
        }
        else {
          std::string const& ed = next_edit_descriptor();
          int n = ed.size();
          if (ed[0] == 'a' && n > 1) {
            n = utils::unsigned_integer_value(ed.data()+1, n-1);
            to_stream(val.elems(), std::min(val.len(), n));
            for(int i=val.len();i<n;i++) {
              to_stream(" ", 1);
            }
          }
          else {
            to_stream(val.elems(), val.len());
          }
        }
        return *this;
      }

      template <typename T, size_t Ndims>
      write_loop&
      operator,(
        cref<T, Ndims> const& val)
      {
        size_t n = val.size_1d();
        T const* val_begin = val.begin();
        for(size_t i=0;i<n;i++) {
          (*this), val_begin[i];
        }
        return *this;
      }

      ~write_loop()
      {
        if (internal_file_len < 0) {
          if (!have_fmt) {
            if (pos == 0) out->put(' ');
          }
          else {
            next_edit_descriptor(/*final*/ true);
          }
          out->put('\n');
          out->flush();
        }
        else {
          if (!have_fmt) {
            throw BOOST_ADAPTBX_NOT_IMPLEMENTED();
          }
          else {
            next_edit_descriptor(/*final*/ true);
            while (pos < internal_file_len) {
              out->put(' ');
              pos++;
            }
          }
        }
      }

    private:

      void
      to_stream(
        char const* buf,
        unsigned n,
        bool space=true)
      {
        if (have_fmt) to_stream_fmt(buf, n);
        else          to_stream_star(buf, n, space);
      }

      void
      to_stream_fmt(
        char const* buf,
        unsigned n)
      {
        out->put(buf, n);
      }

      void
      to_stream_star(
        char const* buf,
        unsigned n,
        bool space=true)
      {
        if (pos == 0) {
          out->put(' ');
          pos = 1;
        }
        else if (pos + (space ? 1 : 0) + n > 80) {
          out->put("\n ", 2);
          pos = 1;
        }
        else if (space) {
          out->put(' ');
          pos++;
        }
        out->put(buf, n);
        pos += n;
      }
  };

  struct common_write
  {
    common& cmn;

    common_write(
      common& cmn_)
    :
      cmn(cmn_)
    {}

    write_loop
    operator()(
      int unit)
    {
      return write_loop(cmn, unit);
    }

    write_loop
    operator()(
      int unit,
      star_type const&)
    {
      return write_loop(cmn, unit, star);
    }

    write_loop
    operator()(
      int const& unit,
      char const* fmt)
    {
      return write_loop(cmn, unit, fmt);
    }

    write_loop
    operator()(
      str_ref const& internal_file,
      star_type const&)
    {
      return write_loop(internal_file, star);
    }

    write_loop
    operator()(
      str_ref const& internal_file,
      char const* fmt)
    {
      return write_loop(internal_file, fmt);
    }
  };

} // namespace fem

#endif // GUARD
